# NFSSVC: root check bypass; use after free due to insufficient locking for NFS upcall arguments

To reproduce the issues, build the attached test.cpp with nfs-entitlement.plist
and run it. Both issues were tested on an ARM virtual machine with macOS 13.0.1.

Issue 1:
The following snippet checks that `NFSSVC_NFSD` or `NFSSVC_ADDSOCK` are only
usable by the super user. But the handlers themselves check if the bit is set.
Therefore simply including an additional unused bit (e.g. NFS_SVC_ADDSOCK | 2)
will pass the equality check but will be handled by NFS_SVC_ADDSOCK. This means
an attacker only needs the `com.apple.private.system-nfssvc` entitlement to
call this syscall.

```c
int
nfssvc(proc_t p __no_nfs_server_unused,
    struct nfssvc_args *uap __no_nfs_server_unused,
    __unused int *retval)
{
    // ...
 /*
  * Must be super user for NFSSVC_NFSD and NFSSVC_ADDSOCK operations.
  */
 if (((uap->flag == NFSSVC_NFSD) || (uap->flag == NFSSVC_ADDSOCK)) && ((error = proc_suser(p)))) {
  return error;
 }

    // ...
 if (uap->flag & NFSSVC_ADDSOCK) {
  if (IS_64BIT_PROCESS(p)) {
   error = copyin(uap->argp, (caddr_t)&user_nfsdarg, sizeof(user_nfsdarg));
  } else {
        // ...
 } else if (uap->flag & NFSSVC_NFSD) {
  error = nfssvc_nfsd();
    // ...
```

Issue 2:
There's a race condition leading to use after free between the `NFSSVC_NFSD`
command and an upcall worker thread.

When a socket is added to the list for nfsd via the `NFSSVC_ADDSOCK` command,
it allocates a `nfsrv_sock` structure to contain a pointer to the original
socket and a newly-allocated `nfsrv_uc_arg` structure which manages the upcall
state.

These sockets are handled by nfsd, which is performed by the `NFSSVC_NFSD`
operation, also exposed explicitly as a syscall. When `nfssvc_nfsd` returns and
no other threads servicing nfsd calls exist, it calls  `nfsrv_cleanup`. This
function calls `nfsrv_zapsock` to teardown the socket data structures created
with `NFSSVC_ADDSOCK` including the upcall. But the upcall is accessed
with locking its `nfsrv_sock` on the upcall queue serviced by the worker threads.
You can see in the diagram below that unrelated locks are used to acquire access
the upcall arguments. Because destruction is an exclusive operation, this is not
safe.

```mermaid
graph TD;
    nfsrv_sock-->|exclusive ns_rwlock|nfsrv_uc_arg;
    queue-->|ucq_lock|nfsrv_uc_arg;
    nfsd-->nfsrv_cleanup;
    nfsrv_cleanup-->|nfsd_mutex|zapsock;
    zapsock-->nfsrv_sock;
    nfsrv_uc_thread-->queue;
```

For additional clarity, I've annotated the output of logs from my fuzzer below.
The fuzzer uses Concurrence, my new deterministic threading library.

```
# First user thread creates the socket
[user_socket]: socket_delegate_wrapper returned error SUCCESS value 0

# NFSD is called from the other user thread
[user_nfsd]: nfssvc (NFSD command): called

# The addsock command spawns several threads to handle socket upcalls in the
# background.
[user_socket]: nfsrv_uc_addsock: starting upcall threads

# Here we show the background upcall thread that we end up interacting with.
# It takes the lock on the queue of pending upcall callbacks.
[nfsrv_uc_thread]: taking ucq_lock 0x7f193898f7d0

# On the user thread, an upcall is allocated for the socket we are adding.
# We then call connectx which will go one to disconnect the socket since it
# is now "connected" to the NFS subsytem.
[user_socket]: nfsrv_uc_addsock: allocated upcall 0x604000042910
[user_socket]: sock_setupcalls_locked: setting upcall for socket 0x61900001f480
[user_socket]: nfssvc (ADDSOCK command) returned error SUCCESS value 0
[user_socket]: connectx called

# The nfsd thread sees that no work is pending and returns. It's the only thread
# calling nfsd, so it cleans up all the added sockets, including the structure
# containg our socket and its upcall callback.
[user_nfsd]: nfsd: last thread is returning, cleaning up on the way out
[user_nfsd]: nfsrv_cleanup: zapping all valid socks
[user_nfsd]: nfsrv_cleanup: found valid slp 0x61300000c280, zapping it

# The socket thread proceeds with the connectx call. As mentioned, it calls
# disconnect which in turn wakes up waiters waiting to receive from the socket.
[user_socket]: connectitx: 0x61900001f480
[user_socket]: soconnectxlocked: 0x61900001f480
[user_socket]: sodisconnectlocked: calling pru_disconnect for 0x61900001f480
[user_socket]: raw_disconnect: calling soisdisconnected 0x61900001f480
[user_socket]: soisdisconnected: waking up receive waiters for socket 0x61900001f480

# Because the socket has an upcall, it calls it to notify the waiter of the
# disconnection. In this case, `nfsrv_uc_proxy` was set as the upcall. This
# function is designated as an upcall proxy because it does not service the
# upcall immediately. The upcall gets enqueued on the upcall queue to be handled
# by the worker thread (nfsrv_uc_thread).
[user_socket]: sowakeup: handling upcall for socket 0x61900001f480
[user_socket]: nfsrv_uc_proxy: enqueuing nfsrv_uc_arg 0x604000042910

# The nfsd thread continues the teardown process. To destroy the upcall context,
# it needs to remove it from any worker queues. Early on, the `nfsrv_uc_thread`
# already acquired the lock so `nfsrc_uc_dequeue` has to wait to lock the upcall
# queue.
[user_nfsd]: sock_setupcalls_locked: clearing upcall for socket 0x61900001f480
[user_nfsd]: nfsrv_zapsock zapping slp 0x61300000c280
[user_nfsd]: nfsrv_uc_dequeue: intend to dequeue nfsrv_uc_arg 0x604000042910
[user_nfsd]: nfsrv_uc_dequeue: waiting on ucq_lock 0x7f193898f7d0

# The worker thread dequeues the worker with its hold on the lock, then drops
# it. The upcall context structure is no longer in the queue, and it now exists
# as a pointer on the stack in the handler function.
[nfsrv_uc_thread]: dequeued nfsrv_uc_arg 0x604000042910
[nfsrv_uc_thread]: dropping ucq_lock 0x7f193898f7d0

# The nfsd thread now acquires the queue lock. It sees the upcall is not currently
# queued as it was already dequeued on the worker thread. Therefore it considers
# it safe to simply free the upcall structure.
[user_nfsd]: nfsrv_uc_dequeue: acquired ucq_lock 0x7f193898f7d0
[user_nfsd]: nfsrv_uc_dequeue: not currently queued
[user_nfsd]: nfsrv_uc_dequeue: freeing upcall 0x604000042910
[user_nfsd]: nfsrv_uc_dequeue: dropping ucq_lock 0x7f193898f7d0

# The stack pointer referencing the upcall is used to call nfsrv_rcv. At this
# point a use-after-free occurs.
[nfsrv_uc_thread]: calling nfsrv_rcv on upcall 0x604000042910
```

This bug is subject to a 90-day disclosure deadline. If a fix for this
issue is made available to users before the end of the 90-day deadline,
this bug report will become public 30 days after the fix was made
available. Otherwise, this bug report will become public at the deadline.
The scheduled deadline is 2023-02-13.
